import {
  IndexFileEntry,
  Resource,
  type Metafile,
  type PackwizIndex,
  type HashFormat,
  doHashesMatch,
} from "@packwizjs/parser";
import { write } from "bun";

function getSaveLocation(
  indexFileEntry: IndexFileEntry,
  index: PackwizIndex,
  metafile?: Metafile,
): Resource {
  const cwd = new Resource(process.cwd());
  const diff = index.location.diff(indexFileEntry.file);
  const fileDirectory = cwd.join(...diff);
  if (metafile) {
    return fileDirectory.parent.join(metafile.filename);
  } else {
    return fileDirectory;
  }
}

async function downloadFile(
  hash: string,
  hashFormat: HashFormat,
  url: Resource,
  path: Resource,
) {
  const response = await fetch(new URL(url.toString()));
  if (!response.ok) {
    throw new Error(`Failed to download ${url}: ${response.statusText}`);
  }
  const arrayBuffer = await response.arrayBuffer();
  if (arrayBuffer.byteLength === 0) {
    console.log(`WARNING: Downloaded file is empty: ${url}`);
  }
  const dataBuffer = Buffer.from(new Uint8Array(arrayBuffer));
  doHashesMatch(hash, hashFormat, dataBuffer);
  console.log(`Saving file to ${path.toString()}`);
  const fileSize = await write(path.toString(), arrayBuffer);
  console.log(`Saved ${fileSize} file to ${path.toString()}`);
}

/**
 * Iterates over the files in a Packwiz index and downloads them concurrently.
 * @param index The Packwiz index to iterate over.
 * @param concurrencyLimit The maximum number of concurrent downloads.
 * @returns A promise that resolves when all files have been downloaded.
 */
export async function iteratePackwizIndex(
  index: PackwizIndex,
  concurrencyLimit: number = 5,
) {
  let currentIndex = 0;
  let activeDownloads = 0;
  const totalFiles = index.files.length;

  return new Promise<void>((resolve, reject) => {
    function downloadNextFile() {
      // Resolve if all file entries are downloaded and no active downloads are ongoing
      if (currentIndex >= totalFiles && activeDownloads === 0) {
        resolve();
        return;
      }

      // If there is still file entries to download and there are empty concurrency slots,
      // start a new download.
      while (activeDownloads < concurrencyLimit && currentIndex < totalFiles) {
        const file = index.files[currentIndex++];
        activeDownloads++;

        (async () => {
          try {
            let saveLocation: Resource;
            let url: Resource;
            let hash: string;
            let hashFormat: HashFormat;

            if (file.metafile) {
              const metafile = await file.parse();
              hash = metafile.provider.hash;
              hashFormat = metafile.provider.hashFormat;
              url = metafile.provider.url;
              saveLocation = getSaveLocation(file, index, metafile);
            } else {
              const diff = index.location.diff(file.file);
              hash = file.hash;
              hashFormat = file.hashFormat;
              url = index.location.parent.join(...diff);
              saveLocation = getSaveLocation(file, index);
            }

            await downloadFile(hash, hashFormat, url, saveLocation);
          } catch (error) {
            console.error(`Error downloading file: ${error}`);
            reject(error);
            return;
          } finally {
            activeDownloads--;
            downloadNextFile();
          }
        })();
      }
    }
    downloadNextFile();
  });
}
